#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <mpi.h>

#define NUMARG 6
#define TAG 8


void usage(char *prog) {
  fprintf(stderr,
"Usage: %s <# of comm (< # of processes)> <max msg len (bytes)> <max iters> <bool: barrier sync at each iter?> <bool: correctness test at each iter?>\n",prog);
  return;
}


main(int argc, char *argv[]) {

int MonNumero;
int n;
int inum, info, msg_rc;
int i,j,q,ii,iinorm,jj;
unsigned char *message;
unsigned char *message_rec;
double T, MaxT;
double time, time_final;
double time1, time2, barrtime, atime();
int taille,source,bufid;
int ntimes;
int nmax, proc, K, barrflag, correctflag;
MPI_Request *request;
MPI_Status  *status;


if(argc < NUMARG) {
   usage(argv[0]);
   exit(-1);
}

MPI_Init(&argc,&argv); /* initialize MPI environment */
MPI_Comm_rank(MPI_COMM_WORLD, &MonNumero); /* get task number */
MPI_Comm_size(MPI_COMM_WORLD,&proc); /* get number of tasks */

request = malloc(2*proc*sizeof(typeof(*request)));
status = malloc(2*proc*sizeof(typeof(*status)));

K=atoi(argv[1]);
if(K>(proc-1)) {
   if(!MonNumero) fprintf(stderr,"Too many comm: %d is the max\n",proc-1);
   MPI_Abort(MPI_COMM_WORLD,-100); exit(-1);
}
nmax=atoi(argv[2]);
if(nmax < 1) {
   if(!MonNumero) fprintf(stderr,"Wrong msg size\n");
   MPI_Abort(MPI_COMM_WORLD,-100); exit(-1);
}
ntimes=atoi(argv[3]);
if(ntimes<1) {
   if(!MonNumero) fprintf(stderr,"Wrong no. of iterations\n");
   MPI_Abort(MPI_COMM_WORLD,-100); exit(-1);
}
barrflag=atoi(argv[4]);
correctflag=atoi(argv[5]);

if(!MonNumero) {
fprintf(stdout,"#============================================\n");
fprintf(stdout,"#   KTK on %d PROCESSORS\n",proc);
fprintf(stdout,"#============================================\n");
}

msg_rc = MPI_Bcast(&K,1,MPI_INT,0,MPI_COMM_WORLD);
if(msg_rc < 0) {  
    fprintf(stderr,"error in broadcast: %d\n",msg_rc);
    MPI_Abort(MPI_COMM_WORLD,-100); /* stop all other tasks */
}
msg_rc = MPI_Bcast(&nmax,1,MPI_INT,0,MPI_COMM_WORLD);
if(msg_rc < 0) {  
    fprintf(stderr,"error in broadcast: %d\n",msg_rc);
    MPI_Abort(MPI_COMM_WORLD,-100); /* stop all other tasks */
}
msg_rc = MPI_Bcast(&ntimes,1,MPI_INT,0,MPI_COMM_WORLD);
if(msg_rc < 0) {  
    fprintf(stderr,"error in broadcast: %d\n",msg_rc);
    MPI_Abort(MPI_COMM_WORLD,-100); /* stop all other tasks */
}
msg_rc = MPI_Bcast(&barrflag,1,MPI_INT,0,MPI_COMM_WORLD);
if(msg_rc < 0) {  
    fprintf(stderr,"error in broadcast: %d\n",msg_rc);
    MPI_Abort(MPI_COMM_WORLD,-100); /* stop all other tasks */
}
msg_rc = MPI_Bcast(&correctflag,1,MPI_INT,0,MPI_COMM_WORLD);
if(msg_rc < 0) {  
    fprintf(stderr,"error in broadcast: %d\n",msg_rc);
    MPI_Abort(MPI_COMM_WORLD,-100); /* stop all other tasks */
}

if(!(message=(char *)malloc(nmax))) {
   fprintf(stderr,"Could not allocate memory %d\n", nmax);
   MPI_Abort(MPI_COMM_WORLD,-100); exit(-1);
}

if(!(message_rec=(char*)malloc(nmax*proc))) {
   fprintf(stderr,"Could not allocate memory %d\n", nmax*proc);
   MPI_Abort(MPI_COMM_WORLD,-100); exit(-1);
}

msg_rc=MPI_Barrier(MPI_COMM_WORLD);

if (!MonNumero) {
  fprintf(stdout,"%10s  %13s  %18s\n",
    "msg size","end2end delay","aggreg. throughput");
  fprintf(stdout,"%10s  %13s  %18s\n",
    "","","(bidirectional)");
  fprintf(stdout,"%10s  %13s  %18s\n",
    "","(usec)","(MByte/s)");
}

time1=MPI_Wtime();
n = 64;
while(n <= nmax)
 {
  barrtime=time=0;
  for(i = 0; i < ntimes; i++) {

    if(barrflag)
       msg_rc=MPI_Barrier(MPI_COMM_WORLD);

    if (correctflag) {
      memset(message_rec,0,n);
      for (j=0;j<n;j++)
        message[j] = (unsigned char)(j*(1+MonNumero));
    }

    time1=MPI_Wtime();

    for(ii = proc-1+MonNumero, jj=0; ii > (proc-1-K)+MonNumero; ii--, jj++) {
      iinorm=ii%proc;
      msg_rc=MPI_Irecv(&message_rec[n*iinorm], n, MPI_BYTE, iinorm,
        TAG, MPI_COMM_WORLD, request+jj);
    }

    for(ii = MonNumero+1; ii < K+MonNumero+1; ii++,jj++) {
      iinorm=ii%proc;
      msg_rc=MPI_Isend(message, n, MPI_BYTE, iinorm,
        TAG, MPI_COMM_WORLD, request+jj);
    }
     
    msg_rc=MPI_Waitall(2*K,request,status);

    time2 = MPI_Wtime();
    MaxT = time2-time1;

    if(barrflag) {
      msg_rc=MPI_Barrier(MPI_COMM_WORLD);
      barrtime+=(MPI_Wtime()-time2);
    }

    MPI_Reduce(&MaxT,&T,1,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);

    if(!MonNumero) time = time + T;

    if (!correctflag) continue;

    for(ii = proc-1+MonNumero; ii > (proc-1-K)+MonNumero; ii--) {
      iinorm=ii%proc;
      for (j=0;j<n;j++)
        if (message_rec[n*iinorm+j] != (unsigned char)(j*(1+iinorm))) {
	  fprintf(stderr,"[%d] err\n",MonNumero);
	  break;
	}
    }

  }

  if(!MonNumero) {
     time_final = time*1e6/ntimes;
 
     fprintf(stdout,"%10d  %13.6f  %18f\n",
       n,(time_final/2.),(double)(n*K*2)/time_final);
#if 0
     if(barrflag) {
        printf("mean time in barrier %f\n",barrtime/ntimes); 
        fprintf(stdout,"time in barrier %f\n",barrtime/ntimes); 
     }
#endif
  }

  n = n*2;
}


msg_rc=MPI_Barrier(MPI_COMM_WORLD);
MPI_Finalize();
exit(0);

}
