#include "mpi.h"
#include <stdlib.h>
#include <stdio.h>
#include "test.h"


int main( int argc, char **argv )
{

    MPI_Comm comm;
    unsigned char    *sbuf, *rbuf;
    int      rank, size, cnk;
    int      sendcounts, recvcounts;
    int      i, j, *p, err, toterr;
    
    MPI_Init( &argc, &argv );
    err = 0;
    
    comm = MPI_COMM_WORLD;

    MPI_Comm_size( comm, &size );
    MPI_Comm_rank( comm, &rank );
    cnk = 1024;
    sbuf = malloc( size * cnk );
    rbuf = malloc( size * cnk );
    if (!sbuf || !rbuf) {
	fprintf( stderr, "Could not allocate buffers!\n" );
	MPI_Abort( comm, 1 );
    }
    
    /* Load up the buffers */
    for (i=0; i<size; i++)
    for (j=0; j<cnk; j++) {
	sbuf[i*cnk+j] = (unsigned char)i;
	rbuf[i*cnk+j] = 0;
    }

    MPI_Alltoall( sbuf, cnk, MPI_BYTE, rbuf, cnk, MPI_BYTE, comm );

    /* Check rbuf */
    for (i=0; i<size*cnk; i++) {
	    if (rbuf[i] != (unsigned char)rank) {
		fprintf( stderr, "[%d] got %d expected %d for %dth\n",
			 rank, rbuf[i],(unsigned char)rank,i);
		err++;
	    }
    }

    free( rbuf );
    free( sbuf );

    MPI_Allreduce( &err, &toterr, 1, MPI_INT, MPI_SUM, MPI_COMM_WORLD );
    MPI_Comm_rank( MPI_COMM_WORLD, &rank );
    if (rank == 0) {
	if (toterr > 0) 
	    fprintf( stderr, "Test FAILED with %d errors\n", toterr );
	else
	    fprintf( stderr, " No Errors\n" );
    }
	
    MPI_Finalize();
    return 0;
}
