/************************************************************************
 *	Read COPYRIGHT file in this directory				*
 ************************************************************************/

/*
 *  $Id: chshort.c,v 1.1.1.1 1997/09/17 20:39:19 gropp Exp $
 *
 *  (C) 1995 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */


/*  ADI protocol: Eager-short  */


#include "mpid.h"
#include "mpiddev.h"
#include "mpimem.h"
#include "reqalloc.h"
#include "queue.h"
#include "chgamma.h"


/* Prototype definitions */

int MPID_CH_Eagerb_send_short
	ANSI_ARGS(( void *, int, int, int, int, int, MPID_Msgrep_t ));
int MPID_CH_Eagerb_isend_short
	ANSI_ARGS(( void *, int, int, int, int, int, MPID_Msgrep_t,
	MPIR_SHANDLE * ));
int MPID_CH_Eagerb_recv_short
	ANSI_ARGS(( MPIR_RHANDLE *, int, void * ));
int MPID_CH_Eagerb_save_short
	ANSI_ARGS(( MPIR_RHANDLE *, int, void *));
int MPID_CH_Eagerb_unxrecv_start_short
	ANSI_ARGS(( MPIR_RHANDLE *, void * ));
void MPID_CH_Eagerb_short_delete
	ANSI_ARGS(( MPID_Protocol * ));


/*
 *  Definitions of the protocol functions
 */


/*
 *  This GAMMA handler is only for ports of data messages (both
 *  short-eager and rendez-vous protocols).
 *  Control messages (REQUEST_SEND, OK_TO_SEND)
 *  are managed by the other handler (see chbrndv.c).
 */
void AG_handler_rec(void)
{

register unsigned long t1,t3,t4,peppe=0;

#ifndef AG_FLOWCTL
if (((AG_head + 1) % AG_MAXQLEN) != AG_tail) {
#endif

AG_src_queue[AG_head].src = gamma_active_port;

/*  Memcpy msg header into the queue of incoming headers.  */
memcpy(&AG_pkt_queue[AG_head].p, AG_buffer[gamma_active_port],
	sizeof(MPID_PKT_HEAD_T));

if (
((MPID_PKT_HEAD_T *)(AG_buffer[gamma_active_port]))->mode == MPID_PKT_SHORT
) {
	/*  Data of short-eager protocol.  */

	/*
	 *  Is this expected or not?
	 *  Try to find a matching receive request in posted queue,
	 *  excluding those ones already marked as "rndv_pending".
	 *  If found, remove the request from posted queue.
	 *  This is similar to MPID_Search_posted_queue() (queue.c).
	 */
	if (MPID_recvs.posted.how_many) {
		t4 = MPID_recvs.posted.head;
		t3 = t4;
		/*  Safety guard.  */
		for (;peppe < AG_MAXQLEN+10;peppe++) {
			if (
(((MPID_PKT_HEAD_T *)(AG_buffer[gamma_active_port]))->context_id)
== MPID_recvs.posted.queue[t4].context_id
&&
(((((MPID_PKT_HEAD_T *)(AG_buffer[gamma_active_port]))->tag)
^ MPID_recvs.posted.queue[t4].tag) & MPID_recvs.posted.queue[t4].tagmask) == 0
&&
(((((MPID_PKT_HEAD_T *)(AG_buffer[gamma_active_port]))->lrank)
^ MPID_recvs.posted.queue[t4].lsrc)& MPID_recvs.posted.queue[t4].srcmask) == 0
&&
(!MPID_recvs.posted.queue[t4].rndv_pending)
			) {
				/*  Message is expected.  */
				AG_src_queue[AG_head].rhandle =
					MPID_recvs.posted.queue[t4].ptr;
				AG_src_queue[AG_head].unexp_entry = -1;
				/*
				 *  AG_src_queue[AG_head].temp_buf
				 *  may be safely left undefined in this case.
				 */
				MPID_recvs.posted.queue[t4].free = 1;
				MPID_recvs.posted.how_many--;
				if (!MPID_recvs.posted.how_many) {
					MPID_recvs.posted.head = 0;
					MPID_recvs.posted.tail = 0;
				}
				else {
					MPID_recvs.posted.queue[t3].next =
					MPID_recvs.posted.queue[t4].next;
					if (t4 == MPID_recvs.posted.head)
					MPID_recvs.posted.head =
					MPID_recvs.posted.queue[t4].next;
					if (t4 == MPID_recvs.posted.tail)
					MPID_recvs.posted.tail = t3;
				}
				if (
((MPID_PKT_HEAD_T *)(AG_buffer[gamma_active_port]))->len == 0
				) {
					/*  No payload.  */
					AG_head = (AG_head + 1) % AG_MAXQLEN;
					gamma_signal(AG_NEWMSG_SEM);
					return;
				}
				/*  There is a payload.  */
				/*
				 *  Setup the payload transfer
				 *  to destination buffer.
				 */
				gamma_switch_to_buffer(
					gamma_active_port,
					MPID_recvs.posted.queue[t4].buf,
					MPID_recvs.posted.queue[t4].len
				);
				AG_head = (AG_head + 1) % AG_MAXQLEN;
				gamma_signal(AG_NEWMSG_SEM);
				return;
			}
			if (t4 == MPID_recvs.posted.tail) break;
			t3 = t4;
			t4 = MPID_recvs.posted.queue[t4].next;
		}
		if (peppe == AG_MAXQLEN+10) {
			/*  Oops!  Inconsistent recv queue!  */
			/* peppe qui perfezionare la segnalazione di errore */
			if (((MPID_PKT_HEAD_T *)
				(AG_buffer[gamma_active_port]))->len
			) {
				/*  Discard payload.  */
				gamma_switch_to_buffer(gamma_active_port,NULL,
								0);
				AG_head = (AG_head + 1) % AG_MAXQLEN;
			}
			gamma_sigerr(AG_NEWMSG_SEM);
			return;
		}
	}

	/*  Message is unexpected.  */
	/*
	 *  Find the first free position of unexpected queue.
	 *  Then, pre-compute the entry by filling up all the
	 *  data relevant to a future receive match.
	 *  The entry is incomplete, since no rhandle
	 *  is allocated here; it will be
	 *  completed by MPID_CH_Check_incoming() (chchkdev.c).
	 *  This is similar to a subset of MPID_Enqueue() (queue.c).
	 */
#ifndef AG_FLOWCTL
	if (MPID_recvs.unexpected.how_many == AG_MAXQLEN) {	
		/* No room in unexpected queue. Message discarded. */
		if (((MPID_PKT_HEAD_T *)(AG_buffer[gamma_active_port]))->len) {
			/*  Discard payload */
			gamma_switch_to_buffer(gamma_active_port,NULL,0);
			AG_head = (AG_head + 1) % AG_MAXQLEN;
		}
		goto ERR;
	}
#endif
	for (t4 = MPID_recvs.unexpected.tail;;t4 = (t4+1) % AG_MAXQLEN)
		if (MPID_recvs.unexpected.queue[t4].free) break;
	/*  We now assume that the above loop provided a value for t4.  */
	if (MPID_recvs.unexpected.how_many) {
		MPID_recvs.unexpected.queue[MPID_recvs.unexpected.tail].next
		= t4;
		MPID_recvs.unexpected.tail = t4;
	}
	MPID_recvs.unexpected.how_many++;
	MPID_recvs.unexpected.queue[t4].free = 0;
	MPID_recvs.unexpected.queue[t4].next = 0;
	/* Store the selection criteria is an easy-to-get place */
	MPID_recvs.unexpected.queue[t4].context_id =
((MPID_PKT_HEAD_T *) (AG_buffer[gamma_active_port]))->context_id;
	MPID_recvs.unexpected.queue[t4].tag =
((MPID_PKT_HEAD_T *) (AG_buffer[gamma_active_port]))->tag;
	MPID_recvs.unexpected.queue[t4].lsrc =
((MPID_PKT_HEAD_T *) (AG_buffer[gamma_active_port]))->lrank;
	MPID_recvs.unexpected.queue[t4].ptr =
(MPIR_RHANDLE *)((unsigned long)-1);
	t3 = ((MPID_PKT_HEAD_T *)(AG_buffer[gamma_active_port]))->len;
	if (t3 == 0) {
		/*  No payload.  */
		AG_src_queue[AG_head].unexp_entry = t4;
		AG_head = (AG_head + 1) % AG_MAXQLEN;
		gamma_signal(AG_NEWMSG_SEM);
		return;
	}
	/*  There is a payload.  */
	/*  Get a free temporary buffer...  */
	t1 = (unsigned long)malloc(t3);
	if (!t1) {
		/* No temporary buffer!!  Discard... */
		if (((MPID_PKT_HEAD_T *)(AG_buffer[gamma_active_port]))->len) {
			/*  Discard payload */
			gamma_switch_to_buffer(gamma_active_port,NULL,0);
			AG_head = (AG_head + 1) % AG_MAXQLEN;
		}
		goto ERR;
	}
	AG_src_queue[AG_head].unexp_entry = t4;
	AG_src_queue[AG_head].temp_buf = (void *)t1;
	/*  Setup the payload transfer to temporary buffer.  */
	gamma_switch_to_buffer(gamma_active_port,(void *)t1,t3);
	AG_head = (AG_head + 1) % AG_MAXQLEN;
	gamma_signal(AG_NEWMSG_SEM);
	return;
}

if (
((MPID_PKT_HEAD_T *)(AG_buffer[gamma_active_port]))->mode == MPID_PKT_LONG
) {
	/*  Data of rendezvous protocol.  */

	/*
	 *  We are sure this is expected (at least, it should be).
	 *  Find the matching receive request in posted queue,
	 *  which has been marked as "rndv_pending".
	 *  When found, remove the request from posted queue.
	 *  This is similar to MPID_Search_posted_queue() (queue.c).
	 */
	if (MPID_recvs.posted.how_many) {
		t4 = MPID_recvs.posted.head;
		t3 = t4;
		/*  Safety guard.  */
		for (;peppe < AG_MAXQLEN+10;peppe++) {
			if (
(((MPID_PKT_HEAD_T *)(AG_buffer[gamma_active_port]))->context_id)
== MPID_recvs.posted.queue[t4].context_id
&&
(((((MPID_PKT_HEAD_T *)(AG_buffer[gamma_active_port]))->tag)
^ MPID_recvs.posted.queue[t4].tag) & MPID_recvs.posted.queue[t4].tagmask) == 0
&&
(((((MPID_PKT_HEAD_T *)(AG_buffer[gamma_active_port]))->lrank)
^ MPID_recvs.posted.queue[t4].lsrc)& MPID_recvs.posted.queue[t4].srcmask) == 0
&&
MPID_recvs.posted.queue[t4].rndv_pending
			) {
				/*  Message is expected.  */
				AG_src_queue[AG_head].rhandle =
					MPID_recvs.posted.queue[t4].ptr;
				AG_src_queue[AG_head].unexp_entry = -1;
				/*
				 *  AG_src_queue[AG_head].temp_buf
				 *  may be safely left undefined in this case.
				 */
				MPID_recvs.posted.queue[t4].free = 1;
				MPID_recvs.posted.how_many--;
                                if (!MPID_recvs.posted.how_many) {
                                        MPID_recvs.posted.head = 0;
                                        MPID_recvs.posted.tail = 0;
                                }
                                else {
                                        MPID_recvs.posted.queue[t3].next =
                                        MPID_recvs.posted.queue[t4].next;
                                        if (t4 == MPID_recvs.posted.head)
                                        MPID_recvs.posted.head =
                                        MPID_recvs.posted.queue[t4].next;
                                        if (t4 == MPID_recvs.posted.tail)
                                        MPID_recvs.posted.tail = t3;
                                }
				if (
((MPID_PKT_HEAD_T *)(AG_buffer[gamma_active_port]))->len == 0
				) {
					/*  No payload.  */
					AG_head = (AG_head + 1) % AG_MAXQLEN;
					gamma_signal(AG_NEWMSG_SEM);
					return;
				}
				/*  There is a payload.  */
				/*
				 *  Setup the payload transfer
				 *  to destination buffer.
				 */
				gamma_switch_to_buffer(
					gamma_active_port,
					MPID_recvs.posted.queue[t4].buf,
					MPID_recvs.posted.queue[t4].len
				);
				AG_head = (AG_head + 1) % AG_MAXQLEN;
				gamma_signal(AG_NEWMSG_SEM);
				return;
			}
			if (t4 == MPID_recvs.posted.tail) break;
			t3 = t4;
			t4 = MPID_recvs.posted.queue[t4].next;
		}
	}
	/*  Should NEVER arrive here!!  */
	if (((MPID_PKT_HEAD_T *)(AG_buffer[gamma_active_port]))->len) {
		/*  Discard payload.  */
		gamma_switch_to_buffer(gamma_active_port,NULL,0);
		AG_head = (AG_head + 1) % AG_MAXQLEN;
	}
/* peppe qui perfezionare la segnalazione di errore */
	gamma_signal(AG_2);
	gamma_sigerr(AG_NEWMSG_SEM);
	return;
}

/*  Should NEVER arrive here!!  */
if (((MPID_PKT_HEAD_T *)(AG_buffer[gamma_active_port]))->len) {
	/*  Discard payload.  */
	gamma_switch_to_buffer(gamma_active_port,NULL,0);
}
AG_src_queue[AG_head].unexp_entry = -1;
AG_head = (AG_head + 1) % AG_MAXQLEN;
gamma_sigerr(AG_NEWMSG_SEM);
return;

#ifndef AG_FLOWCTL
}

/*
 *  Next ctl/short msg would overflow the AG_src_queue.
 *  It is much easier to detect the overflow now.
 *  So we assume that there is overflow, and discard
 *  this message by not notifying it.
 */
#endif
/* peppe qui perfezionare la segnalazione di errore */
ERR:
gamma_sigerr(AG_NEWMSG_SEM);
return;

}


/*  Blocking send.  */


int MPID_CH_Eagerb_send_short(
	void *buf, int len, int src_lrank, int tag, int context_id, int dest,
	MPID_Msgrep_t msgrep
)
{
	MPID_PKT_HEAD_T pkt;

	if (dest != MPID_MyWorldRank) {
#ifdef AG_FLOWCTL
	if (AG_flow[dest].msgs > AG_MAXQLEN/(2*(MPID_MyWorldSize-1))) {
		AG_flow[dest].msgs--;
#endif
		/*
		 *  These references are ordered to match the
		 *  order they appear in the structure.
		 */
		pkt.mode = MPID_PKT_SHORT;
		pkt.lrank = src_lrank;
		pkt.context_id = context_id;
		pkt.tag = tag;
		pkt.len = len;
		MPID_DO_HETERO(pkt.msgrep = (int)msgrep);

		DEBUG_PRINT_SEND_PKT("S Sending",&pkt);

		MPID_PKT_PACK( &pkt, sizeof(MPID_PKT_HEAD_T), dest );
    
		/*  Always use a blocking send for short messages.  */

		DEBUG_PRINT_PKT_DATA("S Getting data from buf",&pkt);
		DEBUG_PRINT_SEND_PKT("S Sending short msg",&pkt);

#if AGS_DEBUG
		printf("[%d] sending %d bytes to port %d\n",
			gamma_my_node(),len,dest);
		fflush(stdout);
#endif
		GAMMASEND_2P(dest,&pkt,sizeof(MPID_PKT_HEAD_T),buf,len);

		DEBUG_PRINT_MSG("S Sent short msg");
		return MPI_SUCCESS;
#ifdef AG_FLOWCTL
	}
	/*  half of msg credits have gone.  use rndv to force credit renewal */
#if AGS_DEBUG
	printf("[%d] msgs %d: switching to rndv send\n",
		gamma_my_node(), AG_flow[dest].msgs);
	fflush(stdout);
#endif
	return
	MPID_CH_Rndvb_send(buf,len,src_lrank,tag,context_id,dest,msgrep);
#if AGS_DEBUG
	printf("[%d] back from rndv send\n", gamma_my_node()); fflush(stdout);
#endif
#endif
	}
	return
	MPID_CH_Rndvb_send(buf,len,src_lrank,tag,context_id,dest,msgrep);
}


/*  Non-blocking send.  */


#ifdef USE_GAMMA_ISEND
static int MPID_CH_Eagerb_tsend_short(MPIR_SHANDLE *shandle)
{
	int res;
	res = gamma_tsend(shandle->is_non_blocking);
	if (res == 0) {
		/*  Operation not yet complete  */
		return -1;
	}
	if (res > 0) {
		/*  Operation completed successfully  */
		shandle->is_complete = 1;
		return 0;
	}
	/*  Operation failed  */
	return MPI_ERR_INTERN;
}


static int MPID_CH_Eagerb_wsend_short(MPIR_SHANDLE *shandle)
{
	int res;
	res = gamma_wsend(shandle->is_non_blocking);
	if (res == 0) {
		/*  Operation completed successfully  */
		shandle->is_complete = 1;
		return MPI_SUCCESS;
	}
	/*  Operation failed  */
	return MPI_ERR_INTERN;
}
#endif


static inline int MPID_CH_Eagerb_isend_short_noflowctl(
	void *buf, int len, int src_lrank, int tag, int context_id, int dest,
	MPID_Msgrep_t msgrep, MPIR_SHANDLE *shandle
)
{
	MPID_PKT_HEAD_T pkt;

	/*
	 *  These references are ordered to match the order they appear in the 
	 *  structure.
	 */
	pkt.mode = MPID_PKT_SHORT;
	pkt.lrank = src_lrank;
	pkt.context_id = context_id;
	pkt.tag = tag;
	pkt.len = len;
	MPID_DO_HETERO(pkt.msgrep = (int)msgrep);

	DEBUG_PRINT_SEND_PKT("S Sending",&pkt);

	MPID_PKT_PACK( &pkt, sizeof(MPID_PKT_HEAD_T), dest );
    
	/*
	Always use a blocking send for short messages.
	(May fail with systems that do not provide adequate
	buffering.  These systems should switch to non-blocking sends)
	*/

	DEBUG_PRINT_PKT_DATA("S Getting data from buf",&pkt);
	DEBUG_PRINT_SEND_PKT("S Sending long msg",&pkt);

	/* In case the message is marked as non-blocking, indicate that we don't
	need to wait on it.  We may also want to use nonblocking operations
	to send the envelopes.... */

#if AGS_DEBUG
	printf("[%d] sending %d bytes to port %d\n",gamma_my_node(),len,dest);
	fflush(stdout);
#endif
#ifdef USE_GAMMA_ISEND
	shandle->is_complete = 0;
	shandle->is_non_blocking = GAMMAISEND_2P(
		dest,&pkt,sizeof(MPID_PKT_HEAD_T),buf,len
	);
	shandle->test = MPID_CH_Eagerb_tsend_short;
	shandle->wait = MPID_CH_Eagerb_wsend_short;
#else
	shandle->is_complete = 1;
	GAMMASEND_2P(
		dest,&pkt,sizeof(MPID_PKT_HEAD_T),buf,len
	);
#endif

	DEBUG_PRINT_MSG("S Sent long msg");
	return MPI_SUCCESS;
}


int MPID_CH_Eagerb_isend_short(
	void *buf, int len, int src_lrank, int tag, int context_id, int dest,
	MPID_Msgrep_t msgrep, MPIR_SHANDLE *shandle
) {

	if (dest != MPID_MyWorldRank) {
#ifdef AG_FLOWCTL
	if (AG_flow[dest].msgs) {
		AG_flow[dest].msgs--;
		/*  If half of msg credits have gone, use rndv  */
		if (AG_flow[dest].msgs <= AG_MAXQLEN/(2*(MPID_MyWorldSize-1))) {
#if AGS_DEBUG
			printf("[%d] msgs %d: using rndv isend\n",
				gamma_my_node(), AG_flow[dest].msgs);
			fflush(stdout);
#endif
			return
			MPID_CH_Rndvb_isend(
				buf,len,src_lrank,tag,context_id,dest,msgrep,
				shandle);
		}
		return
		MPID_CH_Eagerb_isend_short_noflowctl(
			buf,len,src_lrank,tag,context_id,dest,msgrep,
			shandle);
	}

	/*  message credit towards dest exhausted. Failure.  */
	fprintf(stderr,
"[%d] MPID_CH_Eagerb_isend_short(): Msg credit towards #%d exhausted\n",
		gamma_my_node(),dest);
	return MPI_ERR_EXHAUSTED;
#else
	return
	MPID_CH_Eagerb_isend_short_noflowctl(
		buf,len,src_lrank,tag,context_id,dest,msgrep,shandle);
#endif
	}
	/*  ``Send to self''  */
	return
	MPID_CH_Rndvb_isend(
		buf,len,src_lrank,tag,context_id,dest,msgrep,shandle);
}


/*  Blocking recv.  */

int MPID_CH_Eagerb_recv_short(
	MPIR_RHANDLE *rhandle, int from_grank, void *in_pkt
) {
	int msglen;
	int err = MPI_SUCCESS;
    
	msglen = ((MPID_PKT_HEAD_T *)in_pkt)->len;
	/*
	 *  The msg payload has already been copied to destination
	 *  by the GAMMA handler attached to short/ctl ports, and
	 *  has possibly been truncated to fit the recv request.
	 */
	MPID_CHK_MSGLEN(rhandle,msglen,err,in_pkt);
	rhandle->s.count = msglen;
	rhandle->s.MPI_SOURCE = ((MPID_PKT_HEAD_T *)in_pkt)->lrank;
	rhandle->s.MPI_TAG = ((MPID_PKT_HEAD_T *)in_pkt)->tag;
	rhandle->s.MPI_ERROR = err;
	rhandle->is_complete = 1;
	if (rhandle->finish) {
		MPID_DO_HETERO(
			rhandle->msgrep =
			(MPID_Msgrep_t)((MPID_PKT_HEAD_T *)in_pkt)->msgrep
		);
		(rhandle->finish)( rhandle );
	}
	return err;
}


/* Save unexpected message in rhandle */

int MPID_CH_Eagerb_save_short(MPIR_RHANDLE *rhandle, int from, void *in_pkt)
{
	MPID_PKT_HEAD_T *pkt = (MPID_PKT_HEAD_T *)in_pkt;

#if AGS_DEBUG
	printf("[%d] chshort: creating rhandle for unexp msg (temp buf %p)\n",
		gamma_my_node(),AG_tempbuf_ptr);
	fflush(stdout);
#endif
	rhandle->s.count      = pkt->len;
	rhandle->s.MPI_ERROR  = 0;
	rhandle->is_complete  = 1;
	rhandle->from         = from;
	/* Need to save msgrep for heterogeneous systems */
	MPID_DO_HETERO(rhandle->msgrep = (MPID_Msgrep_t)pkt->msgrep);
	/*
	 *  rhandle->start will keep the addr corresponding to
	 *  where this unexp. msg has been buffered.
	 */
	rhandle->start = AG_tempbuf_ptr;
	rhandle->push = MPID_CH_Eagerb_unxrecv_start_short;
	return 0;
}


/*  This is called when it is time to receive an unexpected message.  */

int MPID_CH_Eagerb_unxrecv_start_short(MPIR_RHANDLE *rhandle, void *in_runex)
{
	MPIR_RHANDLE *runex = (MPIR_RHANDLE *)in_runex;
	int msglen, err = 0;

#if AGS_DEBUG
	printf("[%d] chshort: consuming rhandle of unexp msg (temp buf %p)\n",
		gamma_my_node(),runex->start);
	fflush(stdout);
#endif
	msglen = runex->s.count;
	MPID_CHK_MSGLEN(rhandle,msglen,err,NULL);
	/*  At this point we may have msglen < runex->s.count */
	/* Copy the data from the local area and free that area */
	if (runex->s.count > 0) {
		/*
		 *  rhandle->start keeps the pointer
		 *  corresponding to where this unexp. msg has been buffered.
		 */
		memcpy(rhandle->buf,runex->start,msglen);
		free(runex->start);
	}
	MPID_DO_HETERO(rhandle->msgrep = runex->msgrep);

	rhandle->s = runex->s;

	rhandle->s.count = msglen;
	rhandle->s.MPI_ERROR = err;
/* peppe forse inutili o forse necessarie per le persistent ops? */
#if 0
	rhandle->push = 0;
#endif
	rhandle->is_complete = 1;
	if (rhandle->finish)
		(rhandle->finish)( rhandle );
	MPID_RecvFree( runex );
	return err;
}


void MPID_CH_Eagerb_short_delete( p )
MPID_Protocol *p;
{
	FREE( p );
}


MPID_Protocol *MPID_CH_Short_setup()
{
	MPID_Protocol *p;

	p = (MPID_Protocol *) MALLOC( sizeof(MPID_Protocol) );
	if (!p) return 0;
	p->send		= MPID_CH_Eagerb_send_short;
	p->recv		= MPID_CH_Eagerb_recv_short;
	p->isend	= MPID_CH_Eagerb_isend_short;
	p->push_send	= 0;
	p->cancel_send	= 0;
	p->irecv	= 0;
	p->push_recv	= 0;
	p->cancel_recv	= 0;
	p->do_ack	= 0;
	p->unex		= MPID_CH_Eagerb_save_short;
	p->delete	= MPID_CH_Eagerb_short_delete;

	return p;
}
